/*
  inverter program, bring your own 1/2 bridge gate drivers.
  Outputs are pin 9 and pin 10
  Gate drive enable pin 5, logic HIGH
  Fan control, pin 6
  AC voltage set point ADC ch2
  AC output sample, Ch0
  DC volts in, ch5
  Heat sink temp, ch4
  DC current sense, ch3
*/

#include <avr/pgmspace.h>
const char m_0[] PROGMEM = "0 - unused ";  
const char m_1[] PROGMEM = "1 - set low DC volts (";
const char m_2[] PROGMEM = "2 - set high DC volts (";
const char m_3[] PROGMEM = "3 - set low DC current (";
const char m_4[] PROGMEM = "4 - set high DC current (";
const char m_5[] PROGMEM = "5 - set low AC output (";
const char m_6[] PROGMEM = "6 - set high AC output (";  
const char m_7[] PROGMEM = "7 - set AC set point (";
const char m_8[] PROGMEM = "8 - unused";
const char m_9[] PROGMEM = "9 - set AC output frequency (";
const char m_A[] PROGMEM = "A - calibrate 10K NTC (";
const char m_B[] PROGMEM = "B - fan ON temp(";
const char m_C[] PROGMEM = "C - fan OFF temp(";  
const char m_D[] PROGMEM = "D - shutdown temp(";
const char m_E[] PROGMEM = "E - AC output low limit(";
const char m_F[] PROGMEM = "F - LV cutoff voltage(";
const char m_G[] PROGMEM = "G - over current limit(";
const char m_H[] PROGMEM = "H - enable/disable DC LV cutoff (";
const char m_I[] PROGMEM = "I - enable/disable over current cutoff (";  
const char m_J[] PROGMEM = "J - enable/disable AC under volt cutoff (";
const char m_K[] PROGMEM = "K - enable/disable over temp cutoff (";
const char m_L[] PROGMEM = "Z - zero E2PROM";
const char m_M[] PROGMEM = "Y - put defaults into E2PROM";

const char s1[] PROGMEM = "DC low. Enter volts";
const char s2[] PROGMEM = "DC high. Enter volts";
const char s3[] PROGMEM = "DC current low. Enter amps";
const char s4[] PROGMEM = "DC current high. Enter amps";
const char s5[] PROGMEM = "AC output low. Enter volts";
const char s6[] PROGMEM = "AC output high. Enter volts";
const char s7[] PROGMEM = "AC setpoint  Enter volts";
const char s8[] PROGMEM = "unused";
const char s9[] PROGMEM = "AC output frequency (integer, 50 or 60)";
const char sa[] PROGMEM = "Calibrate 10K NTC. Enter temperature";
const char sb[] PROGMEM = "Fan ON temperature (degC)";
const char sc[] PROGMEM = "Fan OFF temperature (degC)";
const char sd[] PROGMEM = "Shut down temperature (degC)";
const char se[] PROGMEM = "AC output Low Limit (percent, e.g. 90)";
const char sf[] PROGMEM = "LV shut down voltage. Enter volts";
const char sg[] PROGMEM = "Over current limit. Enter Amps";
const char sh[] PROGMEM = "DC LV cutoff (1 = enable,0 = disable)";
const char si[] PROGMEM = "Overcurrent cutoff (1 = enable,0 = disable)";
const char sj[] PROGMEM = "AC under volt cutoff (1 = enable,0 = disable)";
const char sk[] PROGMEM = "over temp cutoff (1 = enable,0 = disable)";

char pm_buf[60];  // above progmem ms can not exceed the size of this buffer.

const char * const s_t[] PROGMEM =    
{   
  m_0,  m_1,  m_2,  m_3,  m_4,  m_5,  m_6,  m_7,  m_8,  m_9,  m_A,  m_B,
  m_C,  m_D,  m_E,  m_F,  m_G,  m_H,  m_I,  m_J,  m_K,  m_L,  m_M,  s1,
  s2,  s3,  s4,  s5,  s6,  s7,  s8,  s9,  sa,  sb,  sc,  sd,  se,  sf,
  sg,  sh,  si,  sj,  sk  
 };

#include <EEPROM.h>
#include <Wire.h>
#include <LiquidCrystal_PCF8574.h>
LiquidCrystal_PCF8574 lcd(0x27);

#define ADC_SP  2           // ADC channel for AC setpoint
#define ADC_OUT  0          // channel for AC output sample
#define ADC_DCI 3           // DC current sense
#define ADC_HS  4           // Heat sink temp sensor
#define ADC_DCV 5           // DC supply volts sample


#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define NPWM  200         // 200 for this and 800 for PPWM for 50Hz and 20kHz PWM. Both must be integers.
//#define PPWM  665       // 60.062 Hz - choose one or the other.
#define PPWM  799         // 50.002 Hz

uint16_t overtemp_i, fantemp_i, dc_under_volts_i, ac_under_volts_i,over_current_i;      
float ac_setpoint, ac_output, dcv, dci, hs_temp;
float ac_output_scale;
uint16_t ac_output_offset;

typedef struct  {
float dc_v_low,dc_v_high,
      dc_i_low,dc_i_high,
      ac_v_low,ac_v_high,
      ac_setpoint,
      ntc_temp_cal,fan_on_temp,
      fan_off_temp,over_temp,
      low_limit_pc,lv_cutoff,
      overcurrent;
uint16_t adc_dc_v_low,adc_dc_v_high,
      adc_dc_i_low,adc_dc_i_high,
      adc_ac_v_low,adc_ac_v_high,
      adc_ntc_temp_cal;
uint8_t freq,
      lv_enabled,
      oi_enabled,
      uv_enabled,
      ot_enabled,
      e2prom_ok;
} _nvdata;
      
_nvdata nvd;
      
uint8_t uf,oen,v1low, serial_enabled;
uint16_t pcount;
uint32_t vpwr;
int sst;
uint16_t l[NPWM+1];
float pwr,ch2,ch0;
uint8_t dbounce;
double Input, Output, Setpoint;
double errSum, lastErr,error,dErr;
double kp, ki, kd;
float xv[5],yv[5];

char lcdbuf[82],*lbp,bufc;
String spstr = "";
uint8_t got_data;

float get_float(void);
int get_int(void);


void setup()
  {
  int i;
  float t,u;

  lcdinit();
  
  for(t=0.0,i=0; i <= NPWM; i++,t +=   3.14159/ (float) (NPWM+1))
    {   // 1/2 wave sine lookup table, scaled to (16 bits - 1)
    u = 65535.0 * sin(t);
    l[i] =  (int)u;  
    }
 
  pinMode(9, OUTPUT);     // V1 drive
  pinMode(10, OUTPUT);    // V2 drive
  pinMode(8,INPUT_PULLUP);// on/off button 
  DDRD = 0xff;   // can NOT let arduino system alter any pins from D0 to D7 via calls to pinMode() etc.. It buggers d5 which I need to pull down low-Z upon boot
  noInterrupts();

  TCCR1A = _BV(COM1A1) | _BV(COM1B0) | _BV(COM1B1) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = 10;
  ICR1 = PPWM;
  TIMSK1 |= (1 << TOIE1);
  
  cbi(PORTD,5);   // disable gate drive output. data direction flag set by above statement.

  sst = 0;      // slow start counter 0 - 251
  v1low = 1;    // half wave output flag, 0 or 1
  pcount = 0;   // counter for sine wave lookup table
  uf = 0;       // set to 1 to enable AC output voltage feedback code to execute
  ch2 = ch0 = 0.0;    // set AC voltage feedback and AC voltage setpoint to something. Not needed really.
  pwr = 0.0;    // PWM duty cycle analog, 0.0 to 1.0
  vpwr= 0;      // integer equivalent of pwr, 0 to PPWM
  dbounce = 0;    // debounce counter for input switch
  oen = 0;      // output enable. = 1 to run inverter, = 0 to stop inverter
  
  serial_enabled = 0;  // = 1 if port open, = 0 if closed
  got_data = 0;
  
  get_nvdata();
  
  // set up PID 0.1 0.0 0.01 for biquad LP Fc = 0.01, np=4, sample rate 1000Hz
  kp = 0.1/1000.0;   
  ki = 0.0;
  kd = 0.01/1000.0;
  errSum = 0.0;
  lastErr = 0.0;
  
  interrupts();             // enable all interrupts
  TIMSK0 = 0;               // except arduino clock

  sbi(ADCSRA,ADPS2) ;   // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
  cbi(ADCSRA,ADPS1) ;
  cbi(ADCSRA,ADPS0) ;
  }


ISR(TIMER1_OVF_vect)        
 {    // 20Khz SPWM, code has only 50uS to run, takes about 10uS
  long c;
  
  c = (l[pcount] * vpwr) >> 16;   // scale sin wave by vpwr 32 bit integer calcs.
  if (v1low == 1)
    OCR1B = c;
  else
    OCR1A = c;

  pcount++;
  if(pcount >= NPWM)
    {
    pcount = 0;
    uf=1;
    if (v1low == 1)
      {
       v1low = 0; 
       TCCR1A =  _BV(COM1A1) |  _BV(WGM11);
       sbi(PORTD,7);  // DSO sync pulse
       cbi(PORTD,7);  
      }
    else
      {
       v1low = 1;
       TCCR1A =   _BV(COM1B1) | _BV(WGM11);
      }
    }
}


void do_pid()
{   
   double timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   error = Setpoint - Input;
   errSum += (error * timeChange);
   dErr = (error - lastErr) / timeChange;
   Output = kp * error + ki * errSum + kd * dErr;
   lastErr = error;
}

 
void loop()
{
  float dcw,fch0;
  int a;
  
  if(uf == 1  )    // this will execute at 100Hz
      {         // it runs at near zero volt crossing of AC output     
      sbi(PORTD,2);   
      uf=0;
      

      
      ch0 = (float)analogRead(0)* 0.385;  // get AC output setpoint
      ac_output = ch0;
      
      ac_setpoint = nvd.ac_setpoint;
      dcv = 54.2;
      dci = 20.7;
      hs_temp = 35.0;
      
      check_switch();
      
      if (oen == 1) sst++;   
      if (sst > 251)sst = 251;  
      if (oen == 0) sst--; 
      if (sst <= 0)
        {
          sst = 0;
          cbi(PORTD,5);   // pull IR2184 shutdown LOW to disable gate drive output
          if (serial_enabled == 0)
            {
              Serial.begin(115200);
              serial_enabled = 1;
            }
          if (got_data == 1)
            parse_it();
        }
      else
        {
        sbi(PORTD,5);     // pull it HIGH, to enable output
        if (serial_enabled == 1)
          {
            serial_enabled = 0;
            Serial.end();
          }
        }

      if(sst > 0 && sst < 250)     // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
        {                           // sst can increase or decrease, giving the slow start and slow stop.
        if (oen == 1)             // slow stop? I think of it as de-gauss
          { 
          if(ac_output < ac_setpoint)
            pwr += 1.0/250.0;     // bang-bang control, should AC output reach setpoint prior to PID
          else
            pwr -= 1.0/250.0;
          }
        else
          pwr -= 1.0 /250.0;
        }
        
      if (sst == 251)       // once sst = 251, change to PID. sst will not be incremented further
        {
          
        Setpoint = ac_setpoint;     
        Input = ac_output;  
        do_pid();
        if (Output < -0.05) Output = -0.05;   // ensure some stability during huge transients.
        if (Output > 0.05) Output = 0.05;
        pwr = pwr + Output;

        /*
        if(ac_output < ac_setpoint)
            pwr += 1.0/250.0;     // bang-bang control, should AC output reach setpoint prior to PID
          else
            pwr -= 1.0/250.0;
        */
        }
        
      if (pwr > 1.0) pwr = 1.0;   // clamp. Do math in floating point to prevent any integer overflow
      if (pwr < 0.01) pwr = 0.01; // it's slower. We could do clamp after conversion to int but it's your mosfets...
      vpwr = (int)((float)PPWM * pwr);     



  
      if (oen == 1)
        strncpy(lbp,"ON ",3);
      else
        strncpy(lbp,"OFF",3);
      dcw = dci*dcv;
      sprintf(lbp+40,"%3d",(int)ac_setpoint);*(lbp+43)=' ';
      sprintf(lbp+24,"%3d",(int)ac_output);*(lbp+27)=' ';
      sprintf(lbp+35,"%3d",(int)hs_temp);*(lbp+38)=' ';
      sprintf(lbp+62,"%3d",(int)dci);*(lbp+65)=' ';
      sprintf(lbp+68,"%3d",(int)dcv);*(lbp+71)=' ';
      sprintf(lbp+74,"%5d",(int)dcw);*(lbp+79)='W';
      bufc++;
      poke(lcdbuf[bufc],1);
      if (bufc > 79)
        {
        lcdxy(0,0);
        bufc = 0;
        }
      cbi(PORTD,2);
      }
   
}

void check_switch()
{
  dbounce++;  
  if (dbounce > 100) dbounce = 100;
  if ((PINB & 1) == 0)     // momentary button pressed, pin 8, pulling internal pullup low
    {
    if (dbounce > 80)
      {  
      if(oen == 0)
        oen = 1;    // enable it.
      else
        oen = 0;   // stop inverter
      dbounce=0;
      }       
    }
}


/*
 *  lcd support, custom code for speed. Uses some of the library too
 */
void lcdinit()
  {
  lcd.begin(20,4);
  lcd.clear();
  lbp = &lcdbuf[1];
  strncpy(lbp,    "---                 ",20);
  strncpy(lbp+40, "                    ",20);
  strncpy(lbp+20, "AC      V  Temp    C",20);
  strncpy(lbp+60, "DC    A,    V,     W",20);
  bufc=0;
  }

void lcdxy( uint8_t col, uint8_t row)
  {
  int row_offsets[] = { 0x00, 0x40, 0x14, 0x54   }; 
  poke(0x80 | (col + row_offsets[row]),0);
  }

void poke(uint8_t c,uint8_t cmd)
  {
  uint8_t v2 = c    & 0x0F;
  uint8_t v1 = c >> 4 & 0x0F;
  uint8_t b;
  Wire.beginTransmission(0x27);
  b = (v1 << 4) + 4 + cmd;
  Wire.write(b);
  _delay_us(2);
  b = (v1 << 4) + cmd;
  Wire.write(b);
  _delay_us(37);
  b = (v2 << 4) + 4 + cmd;
  Wire.write(b);
  _delay_us(2);
  b = (v2 << 4) + cmd;
  Wire.write(b);
   _delay_us(37);
  Wire.endTransmission(); 
  }

/*
 *  config menu and serial port stuff. Only active when oen = 0
 */


void show_menu()
  {
  get_nvdata();
  spstr = "";
  got_data=0;
  Serial.println("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
  pm_str(0);Serial.println("");
  pm_str(1); Serial.print(nvd.dc_v_low);Serial.println("V)");
  pm_str(2); Serial.print(nvd.dc_v_high);Serial.println("V)");
  pm_str(3); Serial.print(nvd.dc_i_low);Serial.println("A)");
  pm_str(4); Serial.print(nvd.dc_i_high);Serial.println("A)");
  pm_str(5); Serial.print(nvd.ac_v_low);Serial.println("V)");
  pm_str(6); Serial.print(nvd.ac_v_high);Serial.println("V)");
  pm_str(7); Serial.print(nvd.ac_setpoint);Serial.println("V)");
  pm_str(9); Serial.print(nvd.freq);Serial.println("Hz)");
  pm_str(10); Serial.print(nvd.ntc_temp_cal);Serial.println("degC)");
  pm_str(11); Serial.print(nvd.fan_on_temp);Serial.println("degC)");
  pm_str(12); Serial.print(nvd.fan_off_temp);Serial.println("degC)");
  pm_str(13); Serial.print(nvd.over_temp);Serial.println("degC)");
  pm_str(14); Serial.print(nvd.low_limit_pc);Serial.println("%)");
  pm_str(15); Serial.print(nvd.lv_cutoff);Serial.println("V)");
  pm_str(16); Serial.print(nvd.overcurrent);Serial.println("A)");
  pm_str(17);  if (nvd.lv_enabled == 1)  Serial.print("ON"); else Serial.print("OFF");  Serial.println(")");
  pm_str(18);  if (nvd.oi_enabled == 1)  Serial.print("ON"); else Serial.print("OFF");  Serial.println(")");
  pm_str(19);  if (nvd.uv_enabled == 1)  Serial.print("ON"); else Serial.print("OFF");  Serial.println(")");
  pm_str(20);  if (nvd.ot_enabled == 1)  Serial.print("ON"); else Serial.print("OFF");  Serial.println(")");
  pm_str(22); Serial.println();
  pm_str(21); Serial.println();
  }

void pm_str(int i)
  {
   strcpy_P(pm_buf, (char*)pgm_read_word(&(s_t[i]))); 
   Serial.print(pm_buf);   
  }



void serialEvent() 
{
  char c;
  while (Serial.available()) 
    {
    c = (char)Serial.read();
    if (got_data == 0)    // dont store while still working on previous string
      {
      spstr += c;
      if (c == '\n')
        got_data = 1;
      }
    }
}


/*
  AC voltage set point ADC ch2
  AC output sample, Ch0
  DC volts in, ch5
  Heat sink temp, ch4
  DC current sense, ch3
*/
void parse_it()
  {
  int i;
  char c;
  spstr.toUpperCase();
  c = spstr.charAt(0);
  switch (c)
    {
    case '?':
      show_menu();
      break;
    case 'Z':      // clear e2prom
      Serial.println("clearing e2prom..");
      for ( i = 0 ; i < EEPROM.length() ; i++) 
        EEPROM.write(i, 0);
      Serial.println("done");
      nv();
      break;
    case 'Y':
      do_nvdefaults();
      nv();
      break;
    case '1': // low for ch5
      pm_str(23);
      nvd.dc_v_low = get_float();
      nvd.adc_dc_v_low = analogRead(5);
      nv(); 
      break;
    case '2': //  high for ch5
      pm_str(24);
      nvd.dc_v_high = get_float();
      nvd.adc_dc_v_high = analogRead(5);
      nv();
      break;   
    case '3': //  low for ch3
      pm_str(25);
      nvd.dc_i_low = get_float();
      nvd.adc_dc_i_low = analogRead(3);
      nv();
      break;     
    case '4': //  high for ch3
      pm_str(26);
      nvd.dc_i_high = get_float();
      nvd.adc_dc_i_high = analogRead(3);
      nv();
      break;       
    case '5': //  low for ch0
      pm_str(27);
      nvd.ac_v_low = get_float();
      nvd.adc_ac_v_low = analogRead(0);
      nv();
      break;     
    case '6': //  high for ch0
      pm_str(28);
      nvd.ac_v_high = get_float();
      nvd.adc_ac_v_high = analogRead(0);
      nv();
      break;         
    case '7': //  get AC setpoint in AC volts
      pm_str(29);
      nvd.ac_setpoint = get_float();
      nv();
      break;     
             
    case '9': //  get freq
      pm_str(31);
      nvd.freq = get_int();
      nv();
      break;       
    case 'A': //  cal NTC, ch4
      pm_str(32);
      nvd.ntc_temp_cal = get_float();
      nvd.adc_ntc_temp_cal = analogRead(4);
      nv();
      break;   
    case 'B': //  fan ON temp
      pm_str(33);
      nvd.fan_on_temp = get_float();
      nv();
      break;  
    case 'C': //  fan OFF temp
      pm_str(34);
      nvd.fan_off_temp = get_float();
      nv();
      break;   
    case 'D': //  cut off temp
      pm_str(35);
      nvd.over_temp = get_float();
      nv();
      break;     
    case 'E': //  AC output low
      pm_str(36);
      nvd.low_limit_pc = get_float();
      nv();
      break; 
    case 'F': //  DC input low
      pm_str(37);
      nvd.lv_cutoff = get_float();
      nv();
      break; 
    case 'G': //  over current
      pm_str(38);
      nvd.overcurrent = get_float();
      nv();
      break;
    case 'H': //  enable DC LV
      pm_str(39);
      nvd.lv_enabled = (get_int() > 1);
      nv();
      break; 
    case 'I': //  enable overcurrent
      pm_str(40);
      nvd.oi_enabled = (get_int() > 1);
      nv();
      break; 
    case 'J': //  enable AC UV
      pm_str(41);
      nvd.uv_enabled = (get_int() > 1);
      nv();
      break;     
    case 'K': //  enable over temp
      pm_str(42);
      nvd.ot_enabled = (get_int() > 1);
      nv();
      break;                                                             
    default:   
      spstr="";
      got_data = 0;
  }
 }

void nv()
  {
    EEPROM.put(10,nvd);
    get_nvdata();
    show_menu();
  }

float get_float()
  {
    char c;
    got_data = 0;
    spstr = "";
    while (got_data == 0)
      {
      if (Serial.available()) 
        {
        c = (char)Serial.read();
        spstr += c;
        if (c == '\n')
          got_data = 1;
        }
      }
  return spstr.toFloat();
  }

int get_int()
  {
    char c;
    got_data = 0;
    spstr = "";
    while (got_data == 0)
      {
      if (Serial.available()) 
        {
        c = (char)Serial.read();
        spstr += c;
        if (c == '\n')
          got_data = 1;
        }
      }
  return spstr.toInt();
  }


/*
 read e2prom data, derive local integer equivalents for fast compare to raw ADC values.
 raw ADC data stored at 100 - 150
 floats at 200 - 400
 linear interpolation, 2D, used for all calibration
 initialise progrma with correct data, freq, etc.
 */

void get_nvdata()
{

 EEPROM.get(10,nvd);

 overtemp_i = 100;
 fantemp_i = 100;
 dc_under_volts_i = 100;
 ac_under_volts_i = 100;
 over_current_i = 100;  
}


void do_nvdefaults()
{
 nvd.dc_v_low = 0.1;
 nvd.dc_v_high = 70.0;
 nvd.dc_i_low = 0.2;
 nvd.dc_i_high = 100.0;
 nvd.ac_v_low = 0.3;
 nvd.ac_v_high = 200.0;
 nvd.ac_setpoint = 230.0;
 nvd.freq = 50;   
 nvd.ntc_temp_cal = 35.0;
 nvd.fan_on_temp = 50.0;
 nvd.fan_off_temp = 45.0;
 nvd.over_temp = 80.0;
 nvd.low_limit_pc = 95.0;
 nvd.lv_cutoff = 48.0;
 nvd.overcurrent = 100.0;
 nvd.lv_enabled = 1;
 nvd.oi_enabled = 0;
 nvd.uv_enabled = 0;
 nvd.ot_enabled = 1;
 nvd.e2prom_ok = 42;
 EEPROM.put(10,nvd);
}

